local playerClass = select(2,UnitClass("player"))
if playerClass ~= "ROGUE" and playerClass ~= "WARRIOR" and playerClass ~= "HUNTER" and playerClass ~= "DRUID" and playerClass ~= "PALADIN" and playerClass ~= "SHAMAN" then return end
local playerHybrid
if playerClass == "DRUID" or playerClass == "PALADIN" or playerClass == "SHAMAN" then playerHybrid = true end

if not DrDamage then DrDamage = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0", "AceHook-2.1", "FuBarPlugin-2.0") end
local L = AceLibrary("AceLocale-2.2"):new("DrDamage")
local BS = AceLibrary("Babble-Spell-2.2")
local SEA = AceLibrary("SpecialEvents-Aura-2.0")
local GT = AceLibrary("Gratuity-2.0")
local Deformat = AceLibrary("Deformat-2.0")

--General
local settings
local type = type
local pairs = pairs
local ipairs = ipairs
local tonumber = tonumber
local math_abs = math.abs
local math_floor = math.floor
local math_ceil = math.ceil
local math_min = math.min
local math_max = math.max
local string_match = string.match
local select = select
local UnitLevel = UnitLevel
local Mana_Cost = MANA_COST
local UnitMana = UnitMana

--Melee
local GetCritChance = GetCritChance
local GetCombatRatingBonus = GetCombatRatingBonus
local GetItemInfo = GetItemInfo
local GetInventoryItemLink = GetInventoryItemLink
local GetInventorySlotInfo = GetInventorySlotInfo

local UnitRangedDamage = UnitRangedDamage
local UnitDamage = UnitDamage
local UnitAttackSpeed = UnitAttackSpeed
local UnitAttackPower = UnitAttackPower
local UnitRangedAttackPower = UnitRangedAttackPower

local function DrD_ClearTable(table)
	for k in pairs(table) do
		table[k] = nil
	end
end

local function DrD_Round(x, y)
	return math_floor( x * 10 ^ y + 0.5 ) / 10 ^ y
end

local function DrD_MatchData( data, ... )
	if not data or not ... then
		return false
	end

	if type( data ) == "table" then
		for _, dataName in ipairs( data ) do
			for i = 1, select('#', ...) do
				if dataName == select(i, ...) then
					return true
				end
			end
		end
	else	
		for i = 1, select('#', ...) do
			if data == select(i, ...) then
				return true
			end
		end
	end
	
	return false
end

local function DrD_Set( n, v, setOnly )
	return function(v) 
		settings[n] = v
		if not setOnly and not DrDamage:IsEventScheduled("UpdatingAB") then
			DrDamage:ScheduleEvent("UpdatingAB", DrDamage.UpdateAB, 1.0, DrDamage)
		end
	end
end

if not playerHybrid then
	DrDamage.defaults = {}
end

DrDamage.defaults.HitCalc_M = false

function DrDamage:Melee_Options()
	self:Melee_OnProfileEnable()	
	
	local table
	
	if playerHybrid then
		DrDamage.options.args.Melee = { type = "group", desc = "Melee Options", name = "Melee", order = 3, args = {} }
		table = DrDamage.options.args.Melee.args	
	else
		table = DrDamage.options.args
	end
	
	table.Calculation = {
		type = "group", desc = L["Calculation options"], name = L["Calculation"],
		args = {			
			HitCalc = {			
				type = 'toggle',
				name = L["+Hit calculation"],
				desc = L["Toggles +hit calculation effects into averages on/off"],
				order = 65,
				get = function() return settings["HitCalc_M"] end,
				set = DrD_Set("HitCalc_M"),
			},
		},
	}
end

function DrDamage:Melee_OnEnable()
	self:Melee_InventoryChanged()
	self:Melee_Options()
	self.Melee_Options = nil
	
	self.spellInfo[BS["Attack"]] = { 	[0] = { AutoAttack = true, Melee = true, WeaponDamage = 1, NoNormalization = true },
						["None"] = { 0 },
	}
	
	self.Calculation["Attack"] = function( calculation, _, BuffTalentRanks )
		local mhB, _, _, ohB = GetWeaponEnchantInfo()
		if mhB then
			GT:SetInventoryItem("player", GetInventorySlotInfo("MainHandSlot"))
			local _, _, bName = GT:Find("^([^%(]+) %(%d+ [^%)]+%)$", nil, nil, false, true)
			if bName then
				if string.match( bName, "Windfury %d+" ) then
					local rank = tonumber(string.match(bName, "%d+"))
					local AP = rank and select(rank, 103, 221, 315, 433, 475) --46, 119, 249, 333, 445
					if AP then
						calculation.WindfuryBonus = AP 
						calculation.WindfuryAttacks = 2 * (BuffTalentRanks["Elemental Weapons"] and select(BuffTalentRanks["Elemental Weapons"], 1.13, 1.27, 1.4) or 1)
					end
				elseif string.match( bName, "Windfury Totem %d+" ) then
					local rank = tonumber(string.match(bName, "%d+"))
					local AP = rank and select(rank, 103, 221, 315, 433, 475)
					if AP then
						calculation.WindfuryBonus = AP * (1 + (BuffTalentRanks["Improved Weapon Totems"] and BuffTalentRanks["Improved Weapon Totems"] * 0.15 or 0))
						calculation.WindfuryAttacks = 1
					end				
				end
			end
		end
		if OffhandHasWeapon() and ohB then
			GT:SetInventoryItem("player", GetInventorySlotInfo("SecondaryHandSlot"))
			local _, _, bName = GT:Find("^([^%(]+) %(%d+ [^%)]+%)$", nil, nil, false, true)
			if bName then
				if string.match( bName, "Windfury %d+" ) then
					local rank = tonumber(string.match(bName, "%d+"))
					local AP = rank and select(rank, 103, 221, 315, 433, 475) --46, 119, 249, 333, 445
					if AP then
						calculation.WindfuryBonus_O = AP 
						calculation.WindfuryAttacks_O = 2 * (BuffTalentRanks["Elemental Weapons"] and select(BuffTalentRanks["Elemental Weapons"], 1.13, 1.27, 1.4) or 1)
					end
				elseif string.match( bName, "Windfury Totem %d+" ) then
					local rank = tonumber(string.match(bName, "%d+"))
					local AP = rank and select(rank, 122, 221, 315, 433, 475)
					if AP then
						calculation.WindfuryBonus_O = AP * (1 + (BuffTalentRanks["Improved Weapon Totems"] and BuffTalentRanks["Improved Weapon Totems"] * 0.15 or 0))
						calculation.WindfuryAttacks_O = 1
					end
				end
			end			
		end
	end
	
	self:RegisterEvent("PLAYER_COMBO_POINTS")
	if playerClass == "WARRIOR" then
		self:RegisterBucketEvent("UNIT_RAGE",1)
	end
	if playerClass == "DRUID" then
		self:RegisterBucketEvent("UNIT_ENERGY",1)
	end
end

function DrDamage:Melee_OnProfileEnable()
	settings = self.db.profile
end

local lastRage = 0
function DrDamage:UNIT_RAGE( units )
	for unit in pairs( units ) do
		if unit == "player" then
			if math_abs(UnitMana("player") - lastRage) >= 5 then
				lastRage = UnitMana("player")
				self:CancelScheduledEvent( "UpdatingAB" )
				self:ScheduleEvent("UpdatingAB", self.UpdateAB, 1, self, false, true)
			end
			return
		end
	end
end

local lastEnergy = 0
function DrDamage:UNIT_ENERGY( units )
	for unit in pairs( units ) do
		if unit == "player" then
			if UnitPowerType("player") == 3 and (math_abs(UnitMana("player") - lastEnergy) >= 20 or UnitMana("player") == 100 and lastEnergy ~= 100) then
				lastEnergy = UnitMana("player")
				self:CancelScheduledEvent( "UpdatingAB" )
				self:ScheduleEvent("UpdatingAB", self.UpdateAB, 1, self, false, true)
			end
			return
		end
	end
end

local normalizationTable = {
	["Daggers"] = 1.7,
	["One-Handed Axes"] = 2.4,
	["One-Handed Maces"] = 2.4,
	["One-Handed Swords"] = 2.4,
	["Fist Weapons"] = 2.4,	
	["Bows"] = 2.8,
	["Crossbows"] = 2.8,
	["Guns"] = 2.8,
	["Two-Handed Axes"] = 3.3,
	["Two-Handed Maces"] = 3.3,
	["Two-Handed Swords"] = 3.3,	
	["Polearms"] = 3.3,
	["Staves"] = 3.3,
	["Fishing Pole"] = 3.3,
	["None"] = 2,
}

local oldValues = 0
function DrDamage:Melee_CheckBaseStats()
	local newValues = 
	GetCombatRatingBonus(6)
	+ GetCombatRatingBonus(7)
	+ GetCombatRatingBonus(18)
	+ GetCombatRatingBonus(19)	
	+ self:GetAP() 
	+ self:GetRAP() 
	+ GetCritChance()
	+ GetRangedCritChance()
	+ UnitDamage("player")
	+ select(3,UnitDamage("player"))
	+ UnitRangedDamage("player")
	+ select(2,UnitRangedDamage("player"))

	if newValues ~= oldValues then
		oldValues = newValues
		return true
	end

	return false	
end

local mhSpeed, ohSpeed, rgSpeed
local rgMin, rgMax

function DrDamage:Melee_InventoryChanged()
	local update
	if playerHybrid then
		update = self:CheckRelicSlot()
	end
	if self:Melee_CheckBaseStats() then
		self:RefreshWeaponType()
	
		mhSpeed, ohSpeed = UnitAttackSpeed("player")
		rgSpeed = UnitRangedDamage("player")
		rgMin, rgMax = 0, 0

		if GT:SetInventoryItem("player", GetInventorySlotInfo("MainHandSlot")) then 
			for i = 3, GT:NumLines() do
				local line = GT:GetLine(i,true)
				line = line and tonumber(string_match(line,"%d+\.%d+"))			
				if line then
					mhSpeed = line
					break
				end
			end
		end
		if GT:SetInventoryItem("player", GetInventorySlotInfo("SecondaryHandSlot" )) then
			for i = 3, GT:NumLines() do
				local line = GT:GetLine(i,true)
				line = line and tonumber(string_match(line,"%d+\.%d+"))
				if line then
					ohSpeed = line
					break					
				end
			end
		end
		if GT:SetInventoryItem("player", GetInventorySlotInfo("RangedSlot")) then
			for i = 3, GT:NumLines() do
				local line = GT:GetLine(i,true)
				line = line and tonumber(string_match(line,"%d+\.%d+"))			
				if line then
					rgSpeed = line
					rgMin, rgMax = Deformat(GT:GetLine(i), DAMAGE_TEMPLATE)
					rgMin = (rgMin or 0)
					rgMax = (rgMax or 0)
					break
				end
			end
		end
		return true	
	end
	return update
end

function DrDamage:GetAP()
	local baseAP, posBuff, negBuff = UnitAttackPower("player")
	return baseAP + posBuff + negBuff
end

function DrDamage:GetRAP()
	local baseAP, posBuff, negBuff = UnitRangedAttackPower("player")
	return baseAP + posBuff + negBuff	
end

local mhType, ohType, rgType
function DrDamage:GetWeaponType()
	return mhType, ohType, rgType
end

function DrDamage:RefreshWeaponType()
	local mh = GetInventoryItemLink("player", GetInventorySlotInfo("SecondaryHandSlot"))
	if mh then mhType = select(7,GetItemInfo(mh)) end
	local oh = GetInventoryItemLink("player", GetInventorySlotInfo("RangedSlot"))
	if oh then ohType = select(7,GetItemInfo(oh)) end
	local rg = GetInventoryItemLink("player", GetInventorySlotInfo("MainHandSlot"))
	if rg then rgType = select(7,GetItemInfo(rg)) end
	mhType = mhType or "None"
	ohType = ohType or "None"
	rgType = rgType or "None"
end

function DrDamage:GetRangedBase()
	return rgMin, rgMax
end

function DrDamage:GetWeaponSpeed()
	return mhSpeed, ohSpeed, rgSpeed
end

function DrDamage:GetNormM()
	local mh, oh, rg = normalizationTable[mhType], normalizationTable[ohType], normalizationTable[rgType]
	if not mh or oh or rg then
		self:RefreshWeaponType()
		mh, oh, rg = normalizationTable[mhType], normalizationTable[ohType], normalizationTable[rgType]
	end
	--debug code
	if not mh then
		self:Print( "MH Type: " .. (mhType or "nil") .. " was not found in table. Please tell Gagorian.")
		mh = 2
	end
	if not oh then
		self:Print( "OH Type: " .. (ohType or "nil") .. " was not found in table. Please tell Gagorian.")
		oh = 2
	end
	if not rg then
		self:Print( "RG Type: " .. (rgType or "nil") .. " was not found in table. Please tell Gagorian.")
		rg = 2
	end	
	return mh, oh, rg
end

function DrDamage:WeaponDamage(AP, wspd, ranged, requiresForm)
	local min, max, omin, omax, mod
	local spd, normM, normM_O
	local mh, oh, rg = DrDamage:GetNormM()
	local baseAP = DrDamage:GetAP()
	
	if ranged then
		_, min, max, _, _, mod = UnitRangedDamage("player")
		spd = rgSpeed
		normM = wspd and spd or rg
	else
		min, max, omin, omax, _, _, mod = UnitDamage("player")
		
		if requiresForm then
			return min/mod, max/mod, omin/mod, omax/mod
		end
		
		spd = mhSpeed
		normM = wspd and mhSpeed or mh
		normM_O = wspd and ohSpeed or oh
	end
	
	min = min/mod - (baseAP / 14) * spd + (AP / 14) * normM
	max = max/mod - (baseAP / 14) * spd + (AP / 14) * normM
	
	if normM_O and ohSpeed then
		omin = omin/(0.5*mod) - (baseAP / 14) * ohSpeed + (AP / 14) * normM_O
		omax = omax/(0.5*mod) - (baseAP / 14) * ohSpeed + (AP / 14) * normM_O
	end
	
	return min, max, omin, omax
end

function DrDamage:GetRageGain( avg, calculation )
    	local level = UnitLevel("player")
    	if level >= 60 then
		local avgFactor = 3.5 + 3.5 * (calculation.critPerc/100)
		local conversion = 230.6 + (level-60) * ((274.6-230.6)/10)
		local rage = (avg / conversion * 7.5 + (mhSpeed * avgFactor))/2
		return rage
	else
		return 0
	end
end

local function DrD_GetMeleeHit(ranged, skill, skillO)
	local gear
	if ranged then
		gear = GetCombatRatingBonus(7)
	else
		gear = GetCombatRatingBonus(6)
	end
	
	local hitWeapon = 0
	local hitWeaponO = 0
	local hitWeaponMod = 0
	local lvlDiff, playerLevel, targetLevel = DrDamage:GetLevels()
	
	if UnitName("target") then
		local coeff = UnitIsPlayer("target") and 0.04 or 0.24
		
		--Level difference
		hitWeapon = coeff * 5 * (playerLevel - targetLevel) 
		hitWeaponO = 0.1 * 5 * (playerLevel - targetLevel)
		
		--White melee MH DW penalty modifier
		if not UnitIsPlayer("target") then
			hitWeaponMod = -0.14 * 5 * (playerLevel - targetLevel)
		end
		
		--Weapon skill
		local hitBonus = 0.1 * (skill - playerLevel * 5)
		local hitBonusO = 0.1 * ((skillO or playerLevel * 5) - playerLevel * 5)
		
		--Weapon skill doesn't subtract in pvp
		if UnitIsPlayer("target") then
			hitBonus = math_max(0,hitBonus)
			hitBonusO = math_max(0,hitBonusO)
		end
		
		--Add modifiers
		hitWeapon = hitWeapon + hitBonus
		hitWeaponO = hitWeaponO + hitBonusO
	end
	
	local hit = 95 + gear + hitWeapon
	local hitDW = 86 + gear  + hitWeapon + hitWeaponMod
	local hitDWO = 86 + gear + hitWeaponO
	
	return hit, hitDW - hit, hitDWO - hit
end

local powerTypes = { [0] = "Mana", [1] = "Rage", [2] = "Focus", [3] = "Energy", [4] = "Happiness" }

local ActiveAuras = {}
local BuffTalentRanks = {}
local CalculationResults = {}

local calculation = {}
local CalculationResults = {}

local DrD_DmgCalc, DrD_TalentCalc, DrD_BuffCalc

function DrDamage:MeleeCalc( action, lActionName, tooltip, powerCalc )
	if not action or not lActionName then
		do return end
	end
	
	local actionName = lActionName
	
	if BS and BS:HasReverseTranslation(actionName) then
		actionName = BS:GetReverseTranslation(actionName)
	end
	
	local baseAction = self.spellInfo[lActionName][0]
	local ranged
	calculation.offHdmgM = 0.5
	calculation.bDmgM = 1
	calculation.eDuration = action.eDuration or baseAction.eDuration or 0
	calculation.WeaponDamage = action.WeaponDamage or baseAction.WeaponDamage
	calculation.actionCost = baseAction.Energy or baseAction.Rage
	calculation.critM = 1
	calculation.cooldown = baseAction.Cooldown or 0
	calculation.castTime = action.castTime or baseAction.castTime
	calculation.baseBonus = 0
	calculation.extraDamBonus = 0
	calculation.requiresForm = baseAction.RequiresForm
	
	if not calculation.castTime then
		if baseAction.NextMelee then
			calculation.castTime = mhSpeed
		elseif playerClass == "ROGUE" or playerClass == "DRUID" and GetShapeshiftForm() == 3 then
			calculation.castTime = 1
		else
			calculation.castTime = 1.5
		end
	end
	
	if baseAction.Energy then
		calculation.powerType = 3
	elseif baseAction.Rage then
		calculation.powerType = 1
	else
		calculation.powerType = 0
	end
	
	if type( baseAction.School ) == "table" then
		calculation.actionSchool = baseAction.School[1]
		calculation.actionType = baseAction.School[2]
	else
		calculation.actionSchool = baseAction.School or "Physical"	
	end
	 
	if calculation.actionSchool == "Ranged" then
		ranged = true
		calculation.ranged = true
		calculation.AP = DrDamage:GetRAP()
		calculation.critPerc = GetRangedCritChance()
		calculation.haste = GetCombatRatingBonus(18)
		calculation.dmgM = select(6,UnitRangedDamage("player"))
		calculation.skillMain = UnitRangedAttack("player")
		if baseAction.AutoShot then
			calculation.castTime = UnitRangedDamage("player")
		end
	else
		calculation.AP = DrDamage:GetAP()
		calculation.critPerc = GetCritChance()
		calculation.haste = GetCombatRatingBonus(19)
		calculation.dmgM = select(7,UnitDamage("player"))
		calculation.skillMain, _, calculation.skillOff = UnitAttackBothHands("player")
	end
	
	if calculation.haste == 0 then calculation.haste = 1 end
	calculation.hitPerc, calculation.hitDW, calculation.hitDWO = DrD_GetMeleeHit(ranged, calculation.skillMain, calculation.skillOff)
	
	calculation.minDam = action[1]
	calculation.maxDam = (action[2] or action[1])
	
	if self.RelicSlot then
		if self.RelicSlot[spellName] then
			local data = self.RelicSlot[spellName]
			local count = #data
			if count then
				for i = 1, count - 1, 2 do
					if data[i] and data[i+1] then
						if IsEquippedItem(data[i]) then
							local modType = data["ModType"..((i+1)/2)]
							if not modType then
								calculation.minDam = calculation.minDam + data[i+1]
								calculation.maxDam = calculation.maxDam + data[i+1]
							end
						end
					end
				end
			end
		end
	end	
	
	--Apply talents		
	for talentName, talentRank in pairs(self.talents) do						
		local talentTable = DrDamage.talentInfo[talentName]
		
		for i = 1, #talentTable do
			local talent = talentTable[i]	

			if not talent.Caster then 
				if DrD_MatchData(talent.Spells, actionName, calculation.actionTypeType) or DrD_MatchData(talent.Spells, "All", calculation.actionSchool) then		
					DrD_TalentCalc(baseAction, talentName, talentRank, talent)
				end
			end
		end
	end
	
	for buffName, index, apps, texture, rank in SEA:BuffIter("player") do
		if BS and BS:HasReverseTranslation(buffName) then
			buffName = BS:GetReverseTranslation(buffName)
		end		

		if self.PlayerAura[buffName] then
			local buffData = self.PlayerAura[buffName]

			if DrD_MatchData(buffData.School, calculation.actionSchool) and not baseAction.NoSchoolBuffs
			or not buffData.School and not buffData.Spell
			or DrD_MatchData( buffData.Spell, lActionName )
			or DrD_MatchData( buffData.School, calculation.actionType ) then
				if buffData.ModType == "Special" and self.Calculation[buffName] then
					self.Calculation[buffName]( calculation, ActiveAuras, BuffTalentRanks, index, apps, texture, rank )
				else
					DrD_BuffCalc( buffName, buffData, index, apps, texture, rank, "player", false, baseAction )
				end				
			end
		end		
	end
	for buffName, apps, _, texture, rank, index in SEA:DebuffIter("player") do
		if BS and BS:HasReverseTranslation(buffName) then
			buffName = BS:GetReverseTranslation(buffName)
		end		

		if self.PlayerAura[buffName] then
			local buffData = self.PlayerAura[buffName]

			if DrD_MatchData( buffData.School, calculation.actionSchool ) 
			or not buffData.School and not buffData.Spell					
			or DrD_MatchData( buffData.Spell, lSpellName ) then	
				if buffData.ModType == "Special" and self.Calculation[buffName] then
					self.Calculation[buffName]( calculation, ActiveAuras, BuffTalentRanks, index, apps, texture, rank )
				else
					DrD_BuffCalc( buffName, buffData, index, apps, texture, rank, "player", true, baseAction )
				end
			end
		end
	end	
	for buffName, apps, _, texture, rank, index in SEA:DebuffIter("target") do
		if BS and BS:HasReverseTranslation(buffName) then
			buffName = BS:GetReverseTranslation(buffName)
		end

		if self.Debuffs[buffName] then
			local buffData = self.Debuffs[buffName]			

			if DrD_MatchData( buffData.School, calculation.actionSchool ) and not baseAction.NoSchoolBuffs
			or not buffData.School and not buffData.Spell
			or DrD_MatchData( buffData.School, calculation.actionType )
			or DrD_MatchData( buffData.Spell, lActionName ) then
				if buffData.ModType == "Special" and self.Calculation[buffName] then
					self.Calculation[buffName]( calculation, ActiveAuras, BuffTalentRanks, index, apps, texture, rank )
				else
					DrD_BuffCalc( buffName, buffData, index, apps, texture, rank, "target", true, baseAction )
				end
			end
		end
	end
	
	if self.SetBonuses[playerClass] then
		self.SetBonuses[playerClass]( calculation, ActiveAuras )
	end		
	if self.Calculation[playerClass] then
		self.Calculation[playerClass]( calculation, ActiveAuras, BuffTalentRanks )
	end
	if self.SetBonuses[actionName] then
		self.SetBonuses[actionName]( calculation, ActiveAuras )
	end
	if self.Calculation[actionName] then
		self.Calculation[actionName]( calculation, ActiveAuras, BuffTalentRanks, action )
	end
	
	if baseAction.DualAttack and OffhandHasWeapon() then
		calculation.bDmgM = calculation.bDmgM * 2
	end
	
	local avgTotal = DrD_DmgCalc( actionName, baseAction, action, false, false, tooltip, powerCalc )
	
	if tooltip then
		if not baseAction.NoCrits then
			calculation.critPerc = calculation.critPerc + 1
			CalculationResults.NextCrit = DrD_Round( DrD_DmgCalc( actionName, baseAction, action, true ) - avgTotal, 2 )
			calculation.critPerc = calculation.critPerc - 1	
		end
		if baseAction.APBonus or baseAction.WeaponDamage or baseAction.APGain then
			calculation.AP = calculation.AP + 100
			CalculationResults.NextAP = DrD_Round( DrD_DmgCalc( actionName, baseAction, action, true ) - avgTotal, 2 )
			calculation.AP = calculation.AP - 100
		end
		if not baseAction.Unresistable then
			local temp

			if settings.HitCalc_M then
				temp = avgTotal
			else
				temp = DrD_DmgCalc( actionName, baseAction, action, true, true )
			end

			calculation.hitPerc = calculation.hitPerc + 1
			CalculationResults.NextHit = DrD_Round( DrD_DmgCalc( actionName, baseAction, action, true, true ) - temp, 2 )
		end			
	end	
	
	DrD_ClearTable( BuffTalentRanks )
	DrD_ClearTable( ActiveAuras )	
	DrD_ClearTable( calculation )
	
	return math_floor( avgTotal + 0.5 )
end

DrD_DmgCalc = function( actionName, baseAction, action, nextCalc, hitCalc, tooltip, powerCalc )


	if DrDamage.DmgCalculation[actionName] then
		DrDamage.DmgCalculation[actionName]( calculation, action )
	end

	--Calculation
	local minDam = calculation.minDam * calculation.bDmgM
	local maxDam = calculation.maxDam * calculation.bDmgM
	local minDam_O, maxDam_O
	local minCrit, maxCrit
	local minCrit_O, maxCrit_O	
	local avgHit, avgHit_O
	local avgCrit,avgCrit_O
	local avgTotal
	local avgTotal_O = 0
	local eDuration = calculation.eDuration
	local hitPerc = calculation.hitPerc
	local hits = baseAction.Hits or calculation.Hits
	
	if calculation.WeaponDamage then
		local min, max, min_O, max_O = DrDamage:WeaponDamage( calculation.AP, baseAction.NoNormalization, calculation.ranged, calculation.requiresForm )
		minDam = minDam + (min + calculation.baseBonus) * calculation.WeaponDamage
		maxDam = maxDam + (max + calculation.baseBonus) * calculation.WeaponDamage
		minDam_O = min_O
		maxDam_O = max_O
	end
	if (baseAction.AutoAttack or baseAction.DualAttack) and OffhandHasWeapon() then
		minDam_O = (minDam_O + calculation.baseBonus) * calculation.WeaponDamage * calculation.offHdmgM
		maxDam_O = (maxDam_O + calculation.baseBonus) * calculation.WeaponDamage * calculation.offHdmgM
	end	
	
	if baseAction.ComboPoints then
		local cp = GetComboPoints()
		
		if cp > 0 then
			if action.PerCombo then
				minDam = minDam + action.PerCombo * ( cp - 1 )
				maxDam = maxDam + action.PerCombo * ( cp - 1 )
			end
			if baseAction.ExtraPerCombo then
				minDam = minDam + action.Extra * baseAction.ExtraPerCombo[cp]
				maxDam = maxDam + action.Extra * baseAction.ExtraPerCombo[cp]
			end
			if baseAction.APBonus then
				if type( baseAction.APBonus ) == "table" then
					minDam = minDam + baseAction.APBonus[cp] * calculation.AP
					maxDam = maxDam + baseAction.APBonus[cp] * calculation.AP
				else
					minDam = minDam + baseAction.APBonus * cp * calculation.AP
					maxDam = maxDam + baseAction.APBonus * cp * calculation.AP
				end
			end
			if baseAction.DurationPerCombo then
				eDuration = calculation.eDuration + baseAction.DurationPerCombo * cp
			end
		else
			calculation.zero = true
		end
	else
		if baseAction.APBonus then
			minDam = minDam + baseAction.APBonus * calculation.AP
			maxDam = maxDam + baseAction.APBonus * calculation.AP		
		end
	end
	
	if action.PowerBonus and calculation.powerType == UnitPowerType("player") then
		minDam = minDam + math_max(0, (UnitMana("player") - calculation.actionCost) * action.PowerBonus)
		maxDam = maxDam + math_max(0, (UnitMana("player") - calculation.actionCost) * action.PowerBonus)
	end	
	
	minDam = calculation.dmgM * minDam
	maxDam = calculation.dmgM * maxDam
	
	if baseAction.Weapon and baseAction.Weapon ~= mhType
	or baseAction.RequiresO and baseAction.RequiresO ~= ohType
	or calculation.actionSchool == "Ranged" and rgType == "None"
	or calculation.actionSchool == "Physical" and mhType == "None" and not baseAction.AutoAttack and not baseAction.NoWeapon
	or calculation.zero then
		minDam = 0
		maxDam = 0
	end
	
	avgHit = ((minDam + maxDam) / 2)
	
	if calculation.critPerc > 100 then
		calculation.critPerc = 100
	end	
	
	local critBonus = 0
	local critBonus_O = 0
	
	if not baseAction.NoCrits then
		minCrit = minDam + minDam * calculation.critM
		maxCrit = maxDam + maxDam * calculation.critM
		avgCrit = (minCrit + maxCrit) / 2
		critBonus = (calculation.critPerc / 100) * avgHit * calculation.critM
		avgTotal = avgHit + critBonus
	else
		avgTotal = avgHit
	end
	
	if (baseAction.DualAttack or baseAction.AutoAttack) and OffhandHasWeapon() then
		if baseAction.AutoAttack then
			hitPerc = math_min(hitPerc + calculation.hitDW, 100)
		end
		
		maxDam_O = calculation.dmgM * minDam_O
		maxDam_O = calculation.dmgM * maxDam_O		

		avgHit_O = (minDam_O + maxDam_O)/2
		minCrit_O = minDam_O + minDam_O * calculation.critM
		maxCrit_O = maxDam_O + maxDam_O * calculation.critM
		avgCrit_O = (minCrit_O + maxCrit_O) / 2
		critBonus_O = (calculation.critPerc / 100) * avgHit_O * calculation.critM
		avgTotal_O = avgHit_O + critBonus_O
	end
	
	local extraDam
	local extraDPS = 0
	
	if baseAction.ExtraDamage then
		extraDam = (action.Extra + baseAction.ExtraDamage * calculation.AP) * calculation.dmgM + calculation.extraDamBonus
		avgTotal = avgTotal + extraDam
		extraDPS = extraDam / (baseAction.E_eDuration or baseAction.castTime)
	end
	
	if hits then
		avgTotal = avgTotal + (hits - 1) * avgTotal
	end	
	
	local baseAttack
	if baseAction.NextMelee then
		if baseAction.WeaponDamage then
			local min, max = DrDamage:WeaponDamage( calculation.AP, baseAction.NoNormalization, false, calculation.requiresForm )
			baseAttack = (calculation.dmgM * baseAction.WeaponDamage*(min+max+calculation.baseBonus)/2)
			avgTotal = avgTotal - critBonus - baseAttack
			critBonus = avgTotal * (calculation.critPerc / 100) * calculation.critM
			avgTotal = avgTotal + critBonus
		end
	end
	
	if hitPerc > 100 then
		hitPerc = 100
	end
	
	if not baseAction.Unresistable then
		if settings.HitCalc_M or hitCalc then
			if baseAction.AutoAttack or baseAction.AutoShot then
				avgTotal = (avgTotal - critBonus) * (hitPerc / 100) + critBonus
			else
				avgTotal = avgTotal * (hitPerc / 100)
			end
			if OffhandHasWeapon() then
				if baseAction.AutoAttack then
					avgTotal_O = ( avgTotal_O - critBonus_O ) * (math_min(100, hitPerc + calculation.hitDWO) / 100) + critBonus_O
				elseif baseAction.DualAttack then
					avgTotal_O = avgTotal_O * (hitPerc / 100)
				end
			end
		end
	end
	
	local avgCombined = avgTotal + avgTotal_O
	local DPS, DPSCD, avgWf, avgWf_O, avgTotalWf, avgTotalWf_O

	if baseAction.AutoShot then
		DPS = avgTotal / math_max(0.5, (UnitRangedDamage("player")))
	elseif baseAction.AutoAttack or baseAction.WeaponDPS then
		local spd, ospd = UnitAttackSpeed("player")
		
		if hits then DPS = (avgTotal / (hits - 1)) / spd 
		else DPS = avgTotal /spd end
		
		if ospd then
			DPS = DPS + avgTotal_O / ospd
		end
		
		if baseAction.AutoAttack then
			if calculation.WindfuryBonus then
				local hitmod = calculation.hitPerc/100
				if OffhandHasWeapon() then hitmod = hitmod + calculation.hitDW/100 end
				local min, max = DrDamage:WeaponDamage(calculation.AP)
				avgWf = calculation.dmgM * (calculation.WindfuryAttacks or 1) * ((min+max)/2 + calculation.WindfuryBonus/14)

				avgTotalWf = math_min(1,hitmod) * avgWf + avgWf * calculation.critM * calculation.critPerc/100
				avgCombined = avgCombined + avgTotalWf

				local spd = UnitAttackSpeed("player")
				DPS = DPS + (avgTotalWf * math_min(20,(60/spd)*0.2))/60 --add 3s cd effects later
			end
			if calculation.WindfuryBonus_O then
				local hitmod = math_min(1,(calculation.hitPerc+calculation.hitDWO)/100)
				local _, _, min, max = DrDamage:WeaponDamage(calculation.AP)
				avgWf_O = calculation.dmgM * (calculation.WindfuryAttacks_O or 1) * ((min+max)/2 + calculation.WindfuryBonus_O/14)

				avgWfTotal_O = hitmod * avgWf_O + avgWf_O * calculation.critM * calculation.critPerc/100
				avgCombined = avgCombined + avgWfTotal_O

				local _, spd = UnitAttackSpeed("player")
				DPS = DPS + (avgTotalWf_O * math_min(20,(60/spd)*0.2))/60 --add 3s cd effects later
			end
		end
	else
		if eDuration > 0 then
			DPS = avgCombined / eDuration + extraDPS
		end
		if calculation.cooldown > 0 then
			DPSCD = DrD_Round(avgCombined / calculation.cooldown,1)
		end
	end	
	
	if tooltip then
		DrD_ClearTable( CalculationResults )

		CalculationResults.Avg = 	math_floor(avgHit + 0.5)
		CalculationResults.AvgTotal = 	math_floor(avgTotal + 0.5)
		CalculationResults.Min = 	math_floor(minDam)
		CalculationResults.Max = 	math_ceil(maxDam)
		CalculationResults.Ranged =	ranged
		
		if DPS then
			CalculationResults.DPS = DrD_Round(DPS,1)
		end
		if DPSCD then
			CalculationResults.DPSCD = DPSCD
		end

		if not baseAction.NoCrits then
			CalculationResults.MinCrit = 	math_floor(minCrit)
			CalculationResults.MaxCrit = 	math_ceil(maxCrit)
			CalculationResults.AvgCrit = 	math_floor(avgCrit + 0.5)
			CalculationResults.Crit = 	DrD_Round(calculation.critPerc, 2)
		end

		if avgHit_O then
			CalculationResults.AvgO = 	math_floor(avgHit_O + 0.5)
			CalculationResults.MinO = 	math_floor(minDam_O)
			CalculationResults.MaxO = 	math_ceil(maxDam_O)
			CalculationResults.HitO =	math_ceil(math_min(100, hitPerc + calculation.hitDWO), 2)

			CalculationResults.MinCritO = 	math_floor(minCrit_O)
			CalculationResults.MaxCritO = 	math_ceil(maxCrit_O)
			CalculationResults.AvgCritO = 	math_floor(avgCrit_O + 0.5)
			CalculationResults.AvgTotalO =  math_floor(avgTotal_O + 0.5)
		end
		
		if extraDam then
			CalculationResults.Extra = 	math_floor(extraDam + 0.5)
		end
		
		if avgWf then
			CalculationResults.WindfuryAvg = math_floor(avgWf + 0.5)
			CalculationResults.WindfuryAvgTotal = math_floor(avgTotalWf + 0.5)		
		end
		if avgWf_O then
			CalculationResults.WindfuryAvg_O = math_floor(avgWf + 0.5)
			CalculationResults.WindfuryAvgTotal_O = math_floor(avgTotalWf_O + 0.5)
		end
		
		CalculationResults.Hit = 	DrD_Round(hitPerc, 2)
		CalculationResults.DmgM = 	DrD_Round(calculation.dmgM, 2 )
		CalculationResults.AP = 	calculation.AP
		CalculationResults.Hits = 	hits and " (" .. hits .. ")" or ""
		
		CalculationResults.AvgCombined = math_floor( avgCombined + 0.5 )

		if powerCalc then
			CalculationResults.PowerType = calculation.powerType
		
			if not calculation.actionCost then 
				local manaCost
				if GT:GetLine(2) and Mana_Cost then
					manaCost = Deformat(GT:GetLine(2), Mana_Cost)
				end

				if manaCost then
					manaCost = tonumber(manaCost)
					if manaCost == 0 then
						CalculationResults.DPM = "\226\136\158"
					else
						CalculationResults.DPM = DrD_Round( avgTotal / manaCost, 1 )
					end
				end
			else
				if baseAttack and calculation.powerType == 1 then
					calculation.actionCost = calculation.actionCost + DrDamage:GetRageGain(baseAttack, calculation)
				end
				if calculation.actionCost <= 0 then
					CalculationResults.DPM = "\226\136\158"
				else
					CalculationResults.DPM = DrD_Round(avgTotal / calculation.actionCost, 1)
				end
			end
		end
	end
	
	return avgCombined
end

DrD_TalentCalc = function( baseAction, talentName, talentRank, talent )
	local modType = talent.ModType
	local talentValue

	if type( talent.Effect ) == "table" then
		talentValue = talent.Effect[talentRank]
	else
		talentValue = talent.Effect * talentRank
	end		

	if not modType then
		calculation.dmgM = calculation.dmgM * (1 + talentValue)	
	elseif modType == "Crit" then
		calculation.critPerc = calculation.critPerc + talentValue
	elseif modType == "Hit" then
		calculation.hitPerc = calculation.hitPerc + talentValue
	elseif modType == "BaseDamage" then
		calculation.bDmgM = calculation.bDmgM * (1 + talentValue )
	elseif modType == "Offhand" then
		calculation.offHdmgM = calculation.offHdmgM * (1 + talentValue)
	elseif modType == "PowerCost" then
		if calculation.actionCost then
			calculation.actionCost = calculation.actionCost - talentValue
		end
	elseif modType == "CritMultiplier" then
		calculation.critM = calculation.critM + talentValue
	elseif modType == "SpellDuration" then
		calculation.eDuration = calculation.eDuration + talentValue
	elseif modType == "Cooldown" then
		calculation.cooldown = calculation.cooldown - talentValue	
	elseif modType == "BuffTalentRanks" then
		BuffTalentRanks[talentName] = talentValue
	end

	if modType then
		if DrDamage.Calculation[modType] then DrDamage.Calculation[modType]( calculation, talentValue ) end
	end				
end

DrD_BuffCalc = function( bName, bData, index, apps, texture, rank, target, debuff, baseAction )
	local modType = bData.ModType
	
	if bData.ActiveAura then
		local write
		
		if bData.SelfCast then
			if debuff and select(7,UnitDebuff(target,index)) then
				write = true
			elseif not debuff and select(6,UnitBuff(target,index)) then
				write = true
			end
		end
		if not bData.SelfCast or write then
			if apps and apps > 0 then
				ActiveAuras[bName] = apps
			else
				ActiveAuras[bName] = ( ActiveAuras[bName] or 0 ) + 1
			end
		end
		if modType == "ActiveAura" then return end
	end

	if not bData.Value then return end

	if bData.Ranks then
		if rank then
			rank = tonumber(string_match(rank,"%d+"))
		end
		if not rank then
			rank = bData.Ranks
		end

		if not modType then
			calculation.dmgM = calculation.dmgM * ( 1 + rank * bData.Value )			
		elseif modType == "BaseBonus" then
			calculation.baseBonus = calculation.baseBonus + bData.Value[rank]
		elseif modType == "Windfury" then
			calculation.WindfuryBonus = bData.Value[rank]
		end
	--[[
	elseif bData.Apps then
		if apps and apps > 0 then
			if modType == "Crit" then
				calculation.critPerc = calculation.critPerc + apps * bData.Value
			elseif modType == "BaseDamage" then
				calculation.bDmgM = calculation.bDmgM + apps * bData.Value
			elseif modType == "CastTime" then
				if bData.Value < 0 then
					calculation.castTime = calculation.castTime + apps * bData.Value
				end
			elseif not modType then
				calculation.dmgM = calculation.dmgM * ( 1 + apps * bData.Value )		
			end
		end
	elseif bData.Texture and texture and string.find( texture, bData.Texture ) then
		calculation.dmgM = calculation.dmgM * ( 1 + bData.Value )
	--]]
	elseif modType == "CastTime" then
		if bData.Value < 0 then
			calculation.castTime = calculation.castTime + bData.Value
		else
			if baseAction.NextMelee or baseAction.AutoAttack or baseAction.AutoShot then
				calculation.castTime = calculation.castTime / bData.Value
			end
		end
	elseif modType == "BaseDamage" then
		calculation.bDmgM = calculation.bDmgM + bData.Value
	elseif modType == "Crit" then
		calculation.critPerc = calculation.critPerc + bData.Value
	elseif modType == "Hit" then
		calculation.hitPerc = calculation.hitPerc + bData.Value
	elseif not modType then
		calculation.dmgM = calculation.dmgM * ( 1 + bData.Value )
	end
end

local oldAction
local oldTime = GetTime()

function DrDamage:MeleeTooltip( frame, spellName, baseSpell, tableSpell )
	if spellName ~= oldAction or (GetTime()-oldTime) > 2 then
		oldAction = spellName
		oldTime = GetTime()
		self:MeleeCalc(tableSpell, spellName, true, true)
	end

	frame:AddLine(" ")
	
	if CalculationResults.Name then
		frame:AddLine( CalculationResults.Name, 1, 1, 1 )
		frame:AddLine(" ")
	end	

	local r, g, b = 1, 0.82745098, 0

	if not settings.DefaultColor then
		r, g, b = 0, 0.7, 0
	end

	frame:AddDoubleLine(L["Coeffs:"], CalculationResults.DmgM .. "/" .. CalculationResults.AP, 1, 1, 1, r, g, b  )

	if not baseSpell.noCrits and CalculationResults.Crit then
		frame:AddDoubleLine(L["Crit:"], CalculationResults.Crit .. "%", 1, 1, 1, r, g, b )
	end

	if not baseSpell.Unresistable then
		frame:AddDoubleLine(L["Hit:"], CalculationResults.Hit .. "%", 1, 1, 1, r, g, b )
		if CalculationResults.HitO then
			frame:AddDoubleLine("Off-Hand " .. L["Hit:"], CalculationResults.HitO .. "%", 1, 1, 1, r, g, b )
		end
	end  				

	if not settings.DefaultColor then
		r, g, b = 0.8, 0.8, 0.9
	end
	
	if CalculationResults.AvgO then
		frame:AddLine("Main Hand:")
	end

	frame:AddDoubleLine(L["Avg:"], CalculationResults.Avg .. " (".. CalculationResults.Min .."-".. CalculationResults.Max ..")", 1, 1, 1, r, g, b )

	if CalculationResults.AvgCrit then
		frame:AddDoubleLine(L["Avg Crit:"], CalculationResults.AvgCrit .. " (".. CalculationResults.MinCrit .."-".. CalculationResults.MaxCrit ..")", 1, 1, 1, r, g, b )
	end
	
	if CalculationResults.Extra then
		frame:AddDoubleLine("Additional:", CalculationResults.Extra, 1, 1, 1, r, g, b)
	end
	
	if CalculationResults.WindfuryAvg then
		frame:AddDoubleLine("Windfury Avg/Total:", CalculationResults.WindfuryAvg .. "/" .. CalculationResults.WindfuryAvgTotal, 1, 1, 1, r, g, b)
	end
	
	if CalculationResults.AvgTotal then -- and CalculationResults.AvgTotal > CalculationResults.Avg then
		frame:AddDoubleLine(L["Avg Total:"], CalculationResults.AvgTotal .. CalculationResults.Hits, 1, 1, 1, r, g, b)
	end	
	
	if CalculationResults.AvgO then
		frame:AddLine("Off Hand:")
	end	
	
	if baseSpell.DualAttack or baseSpell.AutoAttack then
		if CalculationResults.AvgO then
			frame:AddDoubleLine(L["Avg:"], CalculationResults.AvgO .. " (".. CalculationResults.MinO .."-".. CalculationResults.MaxO ..")", 1, 1, 1, r, g, b )
			if CalculationResults.AvgCritO then
				frame:AddDoubleLine(L["Avg Crit:"], CalculationResults.AvgCritO .. " (".. CalculationResults.MinCritO .."-".. CalculationResults.MaxCritO ..")", 1, 1, 1, r, g, b )
			end
			if CalculationResults.AvgTotalO then -- and CalculationResults.AvgTotalO > CalculationResults.AvgO then
				frame:AddDoubleLine(L["Avg Total:"], CalculationResults.AvgTotalO, 1, 1, 1, r, g, b  )
			end
			if CalculationResults.WindfuryAvgO then
				frame:AddDoubleLine("Windfury Avg/Total:", CalculationResults.WindfuryAvgO .. "/" .. CalculationResults.WindfuryAvgTotalO, 1, 1, 1, r, g, b)
			end			
		end
	end	
	
	if CalculationResults.AvgTotalO then
		frame:AddLine("-")
		frame:AddDoubleLine("Combined Total:", CalculationResults.AvgCombined, 1, 1, 1, r, g, b  )
	end
	

	if not settings.DefaultColor then
		r, g, b = 0.3, 0.6, 0.5
	end
	
	local bType
	if CalculationResults.Ranged then
		bType = "RAP"
	else
		bType = "AP"
	end
	
	if not baseSpell.NoNext then
		local critA, hitA, apA

		if CalculationResults.NextCrit then
			frame:AddDoubleLine("+1%/" .. self:GetRating("Crit") .. " " .. L["Crit:"], "+" .. CalculationResults.NextCrit, 1, 1, 1, r, g, b )
			if CalculationResults.NextCrit > 0.25 then
				critA = DrD_Round(CalculationResults.AvgCombined * 0.01 / CalculationResults.NextCrit * self:GetRating("Crit", nil, true ), 1 )
			end
		end

		if CalculationResults.NextHit then
			frame:AddDoubleLine("+1%/" .. self:GetRating("MeleeHit") .. " " .. L["Hit:"], "+" .. CalculationResults.NextHit, 1, 1, 1, r, g, b )
			if CalculationResults.NextHit > 0.25 then
				hitA = DrD_Round(CalculationResults.AvgCombined * 0.01 / CalculationResults.NextHit * self:GetRating("MeleeHit", nil, true), 1 )
			end  										
		end

		if CalculationResults.NextAP then
			frame:AddDoubleLine("+100 " .. bType .. ":", "+" .. CalculationResults.NextAP, 1, 1, 1, r, g, b )
			if CalculationResults.NextAP > 0.25 then
				apA = DrD_Round(CalculationResults.AvgCombined / CalculationResults.NextAP, 1)
			end
		end

		frame:AddDoubleLine("+1% Total (Cr/Ht/" .. bType .. "):", (critA or "-") .. "/" .. (hitA or "-") .. "/" .. (apA or "-"), 1, 1, 1, r, g, b )
	end

	if not settings.DefaultColor then
		r, g, b = 0.8, 0.1, 0.1
	end
	
	if not baseSpell.NoDPS then
		if CalculationResults.DPS then
			frame:AddDoubleLine( "DPS:", CalculationResults.DPS, 1, 1, 1, r, g, b )
		end
		if CalculationResults.DPSCD then
			frame:AddDoubleLine( "DPS(CD):", CalculationResults.DPSCD, 1, 1, 1, r, g, b )
		end
	end

	if CalculationResults.DPM and not baseSpell.NoDPM then
		frame:AddDoubleLine( "Dmg/" .. powerTypes[CalculationResults.PowerType], CalculationResults.DPM, 1, 1, 1, r, g, b )
	end

	frame:Show()
end